from outlook_oauth2.core.mailbox import Mailbox

__all__ = ['OutlookAccount']


class OutlookAccount(Mailbox):
    """Sets up access to Outlook account for all methods & classes.

    Attributes:
        access_token: A string OAuth token from Outlook allowing access to a user's account
    """

    def __init__(self, access_token):
        Mailbox.__init__(self, self, 'me')
        self.access_token = access_token

    def get_mailbox(self, email_address):
        """Provides access to another inbox attached to this account"""
        return Mailbox(self, email_address)



from .main import *
from .mailbox import *
from .message import *
from .contact import *
from .folder import *
from .attachment import *

__all__ = ['OutlookAccount', 'Mailbox', 'Message', 'Contact', 'Folder', 'Attachment']



import base64
import logging
import json

from dateutil import parser
import requests

from outlook_oauth2.core.attachment import Attachment
from outlook_oauth2.core.contact import Contact
from outlook_oauth2.core.folder import Folder
from outlook_oauth2.internal.utils import get_valid_filename, check_response

log = logging.getLogger('outlook_oauth2')

__all__ = ['Message']


class Message(object):
    """An object representing an email inside of the OutlookAccount.

        Attributes:
            message_id: A string provided by Outlook identifying this specific email
            body: The body content of the email, including HTML formatting
            body_preview: "The first 255 characters of the body"
            subject: The subject of the email
            sender: The :class:`Contact <outlook_oauth2.core.contact.Contact>` who sent this email. You can set this
                before sending an email to change which account the email comes from (so long as the
                :class:`OutlookAccount <outlook_oauth2.core.main.OutlookAccount>` specified has access to the email.
            to: A list of :class:`Contacts <outlook_oauth2.core.contact.Contact>`. You can also provide a list of strings,
                however these will be turned into :class:`Contacts <outlook_oauth2.core.contact.Contact>` after sending the
                email.
            cc: A list of :class:`Contacts <outlook_oauth2.core.contact.Contact>` in the CC field. You can also provide a
                list of strings, however these will be turned into :class:`Contacts <outlook_oauth2.core.contact.Contact>`
                after sending the email.
            bcc: A list of :class:`Contacts <outlook_oauth2.core.contact.Contact>` in the BCC field. You can also provide a
                list of strings, however these will be turned into :class:`Contacts <outlook_oauth2.core.contact.Contact>`
                after sending the email.
            is_draft: Whether or not the email is a draft.
            importance: The importance level of the email; with 0 indicating low, 1 indicating normal, and 2 indicating
                high. ``Message.IMPORTANCE_LOW``, ``Message.IMPORTANCE_NORMAL``, & ``Message.IMPORTANCE_HIGH`` can be
                used to reference the levels.
            categories: A list of strings, where each string is the name of a category.
            time_created: A datetime representing the time the email was created
            time_sent: A datetime representing the time the email was sent

        """

    IMPORTANCE_LOW = 0
    IMPORTANCE_NORMAL = 1
    IMPORTANCE_HIGH = 2

    def __init__(self, mailbox, body, subject, to_recipients, sender=None,
                 cc=None, bcc=None, message_id=None, **kwargs):
        self.mailbox = mailbox
        self.message_id = message_id

        self.body = body
        self.body_preview = kwargs.get('body_preview', '')
        self.subject = subject

        self.is_draft = kwargs.get('is_draft', None)
        self.importance = kwargs.get('Importance', self.IMPORTANCE_NORMAL)
        self.categories = kwargs.get('categories', [])

        focused = kwargs.get('InferenceClassification', False)
        if focused == 'Focused':
            self._focused = True
        else:
            self._focused = False

        self.sender = sender
        self.to = to_recipients
        self.cc = cc or []
        self.bcc = bcc or []

        self.time_created = kwargs.get('time_created', None)
        self.time_sent = kwargs.get('time_sent', None)

        self._attachments = []
        self._has_attachments = kwargs.get('has_attachments', False)

        self.__is_read = kwargs.get('is_read', False)
        self.__parent_folder_id = kwargs.get('parent_folder_id', None)
        self.__parent_folder = None

    def __str__(self):
        return self.subject

    def __repr__(self):
        return str(self)

    @classmethod
    def _json_to_messages(cls, mailbox, json_value):
        return [cls._json_to_message(mailbox, message) for message in json_value['value']]

    @classmethod
    def _json_to_message(cls, mailbox, api_json):
        uid = api_json['Id']
        subject = api_json.get('Subject', '')

        sender = api_json.get('Sender', {})
        sender = Contact._json_to_contact(sender)

        body = api_json.get('Body', {}).get('Content', '')
        body_preview = api_json.get('BodyPreview', '')

        to_recipients = api_json.get('ToRecipients', [])
        to_recipients = Contact._json_to_contacts(to_recipients)

        is_read = api_json['IsRead']
        has_attachments = api_json['HasAttachments']

        time_created = api_json.get('CreatedDateTime', None)
        if time_created is not None:
            time_created = parser.parse(time_created)

        time_sent = api_json.get('SentDateTime', None)
        if time_sent is not None:
            time_sent = parser.parse(time_sent)

        parent_folder_id = api_json.get('ParentFolderId', None)
        is_draft = api_json.get('IsDraft', None)
        importance = api_json.get('Importance', cls.IMPORTANCE_NORMAL)

        categories = api_json.get('Categories', [])

        return_message = Message(mailbox, body, subject, to_recipients, sender=sender, message_id=uid, is_read=is_read,
                                 time_created=time_created, time_sent=time_sent, parent_folder_id=parent_folder_id,
                                 is_draft=is_draft, importance=importance, body_preview=body_preview,
                                 categories=categories, has_attachments=has_attachments)
        return return_message

    @property
    def focused(self):
        """ Sets and retrieves the 'Focused' status of a Message. If a user has the 'Focused' inbox, messages are
        divided between the Focused and Other tabs. """
        return self._focused

    @focused.setter
    def focused(self, value):

        if not isinstance(value, bool):
            raise TypeError('Message.focused must be a boolean value')

        endpoint = "https://outlook.office.com/api/v2.0/" + self.mailbox.route + "/messages('{}')".format(self.message_id)

        if value:
            data = dict(InferenceClassification='Focused')
        else:
            data = dict(InferenceClassification='Other')

        r = requests.patch(endpoint, data=json.dumps(data), headers=self.mailbox._headers)

        if check_response(r):
            self._focused = value

    @property
    def attachments(self):
        # type: () -> [Attachment]
        if not self._has_attachments:
            return []

        if self._attachments:
            return self._attachments

        endpoint = 'https://outlook.office.com/api/v2.0/' + self.mailbox.route + '/messages/{}/attachments'.format(self.message_id)
        r = requests.get(endpoint, headers=self.mailbox._headers)

        if check_response(r):
            data = r.json()
            self._attachments = Attachment.json_to_attachments(self.mailbox, data)

        return self._attachments

    @property
    def is_read(self):
        """ Set and retrieve the 'Read' status of an email

            >>> message = Message()
            >>> message.is_read
            >>> False
            >>> message.is_read = True
        """
        return self.__is_read

    @is_read.setter
    def is_read(self, boolean):
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.mailbox.route + '/messages/{}'.format(self.message_id)
        payload = dict(IsRead=boolean)

        self._make_api_call('patch', endpoint, data=json.dumps(payload))
        self.__is_read = boolean

    @property
    def parent_folder(self):
        # type: () -> Folder
        """ Returns the :class:`Folder <outlook_oauth2.core.folder.Folder>` this message is in

            >>> from outlook_oauth2 import *
            >>> mailbox = OutlookAccount('')
            >>> message = mailbox.get_messages()[0]
            >>> message.parent_folder
            Inbox
            >>> message.parent_folder.unread_count
            19

        Returns: :class:`Folder <outlook_oauth2.core.folder.Folder>`

        """
        if self.__parent_folder is None:
            self.__parent_folder = self.mailbox.get_folder_by_id(self.__parent_folder_id)

        return self.__parent_folder

    def api_representation(self, content_type):
        """ Returns the JSON representation of this message required for making requests to the API.

        Args:
            content_type (str): Either 'HTML' or 'Text'
        """
        payload = dict(Subject=self.subject, Body=dict(ContentType=content_type, Content=self.body))

        if self.sender is not None:
            payload.update(From=self.sender.api_representation())

        # A list of strings can also be provided for convenience. If provided, convert them into Contacts
        if any(isinstance(item, str) for item in self.to):
            self.to = [Contact(email=email) for email in self.to]

        # Turn each contact into the JSON needed for the Outlook API

        recipients = [contact.api_representation() for contact in self.to]

        payload.update(ToRecipients=recipients)

        # Conduct the same process for CC and BCC if needed
        if self.cc:
            if any(isinstance(email, str) for email in self.cc):
                self.cc = [Contact(email) for email in self.cc]

            cc_recipients = [contact.api_representation() for contact in self.cc]
            payload.update(CcRecipients=cc_recipients)

        if self.bcc:
            if any(isinstance(email, str) for email in self.bcc):
                self.bcc = [Contact(email) for email in self.bcc]

            bcc_recipients = [contact.api_representation() for contact in self.bcc]
            payload.update(BccRecipients=bcc_recipients)

        if self._attachments:
            payload.update(Attachments=[attachment.api_representation() for attachment in self._attachments])

        payload.update(Importance=str(self.importance))

        return dict(Message=payload)

    def _make_api_call(self, http_type, endpoint, extra_headers = None, data=None):
        # type: (str, str, dict, Any) -> None
        """
        Internal method to handle making calls to the Outlook API and logging both the request and response
        Args:
            http_type: (str) 'post' or 'delete'
            endpoint: (str) The endpoint the request will be made to
            headers: A dict of headers to send to the requests module in addition to Authorization and Content-Type
            data: The data to provide to the requests module

        Raises:
            MiscError: For errors that aren't a 401
            AuthError: For 401 errors

        """

        headers = {"Authorization": "Bearer " + self.mailbox.access_token, "Content-Type": "application/json"}

        if extra_headers is not None:
            headers.update(extra_headers)

        log.debug('Making Outlook API request for message (ID: {}) with Headers: {} Data: {}'
                  .format(self.message_id, headers, data))

        if http_type == 'post':
            r = requests.post(endpoint, headers=headers, data=data)
        elif http_type == 'delete':
            r = requests.delete(endpoint, headers=headers)
        elif http_type == 'patch':
            r = requests.patch(endpoint, headers=headers, data=data)
        else:
            raise NotImplemented

        check_response(r)

    def send(self, content_type='HTML'):
        """ Takes the recipients, body, and attachments of the Message and sends.

        Args:
            content_type: Can either be 'HTML' or 'Text', defaults to HTML.

        """

        payload = self.api_representation(content_type)

        endpoint = 'https://outlook.office.com/api/v1.0/' + self.mailbox.route + '/sendmail'
        self._make_api_call('post', endpoint=endpoint, data=json.dumps(payload))

    def forward(self, to_recipients, forward_comment=None):
        # type: (list, str) -> None
        """Forward Message to recipients with an optional comment.

        Args:
            to_recipients: A list of :class:`Contacts <outlook_oauth2.core.contact.Contact>` to send the email to.
            forward_comment: String comment to append to forwarded email.

        Examples:
            >>> john = Contact('john.doe@domain.com')
            >>> betsy = Contact('betsy.donalds@domain.com')
            >>> email = Message()
            >>> email.forward([john, betsy])
            >>> email.forward([john], 'Hey John')
        """
        payload = dict()

        if forward_comment is not None:
            payload.update(Comment=forward_comment)

        # A list of strings can also be provided for convenience.
        # If provided, convert them into Contacts
        if any(isinstance(recipient, str) for recipient in to_recipients):
            to_recipients = [Contact(email=email) for email in to_recipients]

        # Contact() will handle turning itself into the proper JSON format for the API
        to_recipients = [contact.api_representation() for contact in to_recipients]

        payload.update(ToRecipients=to_recipients)

        endpoint = 'https://outlook.office.com/api/v2.0/' + self.mailbox.route + '/messages/{}/forward'.format(self.message_id)

        self._make_api_call('post', endpoint=endpoint, data=json.dumps(payload))

    def reply(self, reply_comment):
        """Reply to the Message.

        Notes:
            HTML can be inserted in the string and will be interpreted properly by Outlook.

        Args:
            reply_comment: String message to send with email.

        """
        payload = '{ "Comment": "' + reply_comment + '"}'
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.mailbox.route + '/messages/' + self.message_id + '/reply'

        self._make_api_call('post', endpoint, data=payload)

    def reply_all(self, reply_comment):
        # type: (str) -> None
        """Replies to everyone on the email, including those on the CC line.

        With great power, comes great responsibility.

        Args:
            reply_comment: The string comment to send to everyone on the email.

        """
        payload = '{ "Comment": "' + reply_comment + '"}'
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.mailbox.route + '/messages/{}/replyall'.format(self.message_id)

        self._make_api_call('post', endpoint, data=payload)

    def delete(self):
        """Deletes the email"""
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.mailbox.route + '/messages/' + self.message_id
        self._make_api_call('delete', endpoint)

    def _move_to(self, destination):
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.mailbox.route + '/messages/' + self.message_id + '/move'
        payload = '{ "DestinationId": "' + destination + '"}'
        r = requests.post(endpoint, data=payload, headers=self.mailbox._headers)
        check_response(r)
        data = r.json()
        self.message_id = data.get('Id', self.message_id)

    def move_to_inbox(self):
        """Moves the email to the account's Inbox"""
        self._move_to('Inbox')

    def move_to_deleted(self):
        """Moves the email to the account's Deleted Items folder"""
        self._move_to('DeletedItems')

    def move_to_drafts(self):
        """Moves the email to the account's Drafts folder"""
        self._move_to('Drafts')

    def move_to(self, folder):
        # type: (Folder) -> None
        """Moves the email to the folder specified by the folder parameter.

        Args:
            folder: A string containing the folder ID the message should be moved to, or a Folder instance

        """
        if isinstance(folder, Folder):
            self.move_to(folder.id)
        else:
            self._move_to(folder)

    def _copy_to(self, destination):
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.mailbox.route + '/messages/' + self.message_id + '/copy'
        payload = '{ "DestinationId": "{}"}'.format(destination)

        self._make_api_call('post', endpoint, data=payload)

    def copy_to_inbox(self):
        """Copies Message to account's Inbox"""
        self._copy_to('Inbox')

    def copy_to_deleted(self):
        """Copies Message to account's Deleted Items folder"""
        self._copy_to('DeletedItems')

    def copy_to_drafts(self):
        """Copies Message to account's Drafts folder"""
        self._copy_to('Drafts')

    def copy_to(self, folder_id):
        # type: (str) -> None
        """Copies the email to the folder specified by the folder_id.

        The folder id must match the id provided by Outlook.

        Args:
            folder_id: A string containing the folder ID the message should be copied to

        """
        self._copy_to(folder_id)

    def attach(self, file_bytes, file_name):
        """Adds an attachment to the email. The filename is passed through Django's get_valid_filename which removes
        invalid characters. From the documentation for that function:

        >>> get_valid_filename("john's portrait in 2004.jpg")
        'johns_portrait_in_2004.jpg'

        Args:
            file_bytes: The bytes of the file to send (if you send a string, ex for CSV, outlook_oauth2 will attempt to
                convert that into bytes before base64ing the content).
            file_name: The name of the file, as a string and leaving out the extension, that should be sent

        """
        try:
            file_bytes = base64.b64encode(file_bytes)
        except TypeError:
            file_bytes = base64.b64encode(bytes(file_bytes, 'utf-8'))

        self._attachments.append(
            Attachment(get_valid_filename(file_name), file_bytes.decode('utf-8'))
        )

    def add_category(self, category_name):
        # type: (str) -> None
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.mailbox.route + '/messages/{}'.format(self.message_id)
        self.categories.append(category_name)
        self._make_api_call('patch', endpoint, data=json.dumps(dict(Categories=self.categories)))



import re

from outlook_oauth2.internal.errors import AuthError, RequestError, APIError


def get_valid_filename(s):
    """
    Shamelessly taken from Django.
    https://github.com/django/django/blob/master/django/utils/text.py

    Return the given string converted to a string that can be used for a clean
    filename. Remove leading and trailing spaces; convert other spaces to
    underscores; and remove anything that is not an alphanumeric, dash,
    underscore, or dot.
    >>> get_valid_filename("john's portrait in 2004.jpg")
    'johns_portrait_in_2004.jpg'
    """
    s = str(s).strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)


def get_response_data(response):
    """ Handles getting response data from the requests module where .json() can raise an error """
    try:
        return response.json()
    except ValueError:
        return response.content


def check_response(response):
    """ Checks that a response is successful, raising the appropriate Exceptions otherwise. """
    status_code = response.status_code

    if 100 < status_code < 299:
        return True

    elif status_code == 401 or status_code == 403:
        message = get_response_data(response)
        raise AuthError('Access Token Error, Received ' + str(status_code) +
                        ' from Outlook REST Endpoint with the message: {}'.format(message))

    elif status_code == 400:
        message = get_response_data(response)
        raise RequestError('The request made to the Outlook API was invalid. Received the following message: {}'.
                           format(message))
    else:
        message = get_response_data(response)
        raise APIError('Encountered an unknown error from the Outlook API: {}'.format(message))



class OutlookError(Exception):
    def __init__(self, value=None):
        self.value = value


class APIError(OutlookError):
    """ Any error received from the Outlook API """


class AuthError(APIError):
    """ Raised when Outlook returns a 401 or 403, generally caused by expired or invalid access tokens """
    def __init__(self, value=None):
        if value is None:
            self.value = 'Access Token Error, double check your access token.'
        else:
            self.value = value


class RequestError(APIError):
    """ Raised from 400 response status codes"""
    def __init__(self, value=None):
        self.value = value


class MiscError(OutlookError):
    def __init__(self, value):
        self.value = value



import json
import requests

from outlook_oauth2.internal.utils import check_response

__all__ = ['Contact']


class Contact(object):
    """ Represents someone sending or receiving an email. Cuts down on the amount of dictionaries floating around that
    each hold the API's syntax and allows for functionality to be added in the future.

    Args:
        email (str): The email of the user
        name (str): The user's name, which is not always provided by the API.

    Keyword Args:
        focused: Whether messages from this sender are always sent to the Focused inbox, or to the Other tab.
            This value is set when retrieving a contact from the API, or after setting it via
            :func:`set_focused() <outlook_oauth2.core.contact.Contact.set_focused>`

    Attributes:
        email: The email of the user
        name: The name of the user
        focused: A boolean indicating whether this contact has an override for their messages to go to the Focused inbox
            or the "Other" inbox. None indicates that the value has not yet been retrieved by the API or set.
    """

    def __init__(self, email, name=None, focused=None):
        # type: (str, str, bool) -> None
        self.email = email
        self.name = name
        self.focused = focused

    def __str__(self):
        if self.name is None:
            return self.email
        return '{} ({})'.format(self.name, self.email)

    def __repr__(self):
        return str(self)

    @classmethod
    def _json_to_contact(cls, json_value):
        contact = json_value.get('EmailAddress', None)
        # The API returns this information in a different format if it's related to Focused inbox overrides
        contact_override = json_value.get('SenderEmailAddress', None)
        if contact is not None:
            email = contact.get('Address', None)
            name = contact.get('Name', None)

            return Contact(email, name)
        # This contains override information
        elif contact_override is not None:
            # Whether they are 'Focused' or 'Other'
            classification = json_value.get('ClassifyAs', 'Other')
            focused = True if classification == 'Focused' else False

            email = contact_override.get('Address', None)
            name = contact_override.get('Name', None)

            return Contact(email, name, focused=focused)
        else:
            return None

    @classmethod
    def _json_to_contacts(cls, json_value):
        # Sometimes, multiple contacts will be provided behind a dictionary with 'value' as the key
        try:
            json_value = json_value['value']
        except TypeError:
            pass
        return [cls._json_to_contact(contact) for contact in json_value]

    def api_representation(self):
        """ Returns the JSON formatting required by Outlook's API for contacts """
        return dict(EmailAddress=dict(Name=self.name, Address=self.email))

    def set_focused(self, mailbox, is_focused):
        # type: (Mailbox, bool) -> bool
        """ Emails from this contact will either always be put in the Focused inbox, or always put in Other, based on
        the value of is_focused.

        Args:
            mailbox (Mailbox): The :class:`Mailbox <outlook_oauth2.core.mailbox.Mailbox>`
                the override should be set for
            is_focused (bool): Whether this contact should be set to Focused, or Other.

        Returns:
            True if the request was successful
        """
        endpoint = 'https://outlook.office.com/api/v2.0/' + mailbox.route + '/InferenceClassification/Overrides'

        if is_focused:
            classification = 'Focused'
        else:
            classification = 'Other'

        data = dict(ClassifyAs=classification, SenderEmailAddress=dict(Address=self.email))

        r = requests.post(endpoint, headers=mailbox._headers, data=json.dumps(data))

        # Will raise an error if necessary, otherwise returns True
        result = check_response(r)

        self.focused = is_focused

        return result



import requests

from outlook_oauth2.internal.utils import check_response

__all__ = ['Folder']


class Folder(object):
    """An object representing a Folder in the OutlookAccount provided.

    Attributes:
        account: The :class:`OutlookAccount <outlook_oauth2.core.main.OutlookAccount>` this folder should be associated with
        id: The static id generated by Outlook to identify this folder.
        folder_name: The name of this folder as displayed in the account
        parent_id: The id of the folder which houses this Folder object
        child_folder_count: The number of child folders inside this Folder
        unread_count: The number of unread messages inside this Folder
        total_items: A sum of all items inside Folder

    """
    def __init__(self, account, folder_id, folder_name, parent_id, child_folder_count, unread_count, total_items):
        self.account = account
        self.parent_id = parent_id
        self.child_folder_count = child_folder_count
        self.unread_count = unread_count
        self.total_items = total_items
        self.name = folder_name
        self.id = folder_id

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)

    @property
    def headers(self):
        return {"Authorization": "Bearer " + self.account.access_token, "Content-Type": "application/json"}

    @classmethod
    def _json_to_folder(cls, account, json_value):
        return Folder(account, json_value['Id'], json_value['DisplayName'], json_value['ParentFolderId'],
                      json_value['ChildFolderCount'], json_value['UnreadItemCount'], json_value['TotalItemCount'])

    @classmethod
    def _json_to_folders(cls, account, json_value):
        return [cls._json_to_folder(account, folder) for folder in json_value['value']]

    def rename(self, new_folder_name):
        """Renames the Folder to the provided name.

        Args:
            new_folder_name: A string of the replacement name.

        Raises:
            AuthError: Raised if Outlook returns a 401, generally caused by an invalid or expired access token.

        Returns:
            A new Folder representing the folder with the new name on Outlook.

        """
        headers = self.headers
        endpoint = 'https://outlook.office.com/api/v2.0' + self.account.route + 'MailFolders/' + self.id
        payload = '{ "DisplayName": "' + new_folder_name + '"}'

        r = requests.patch(endpoint, headers=headers, data=payload)

        if check_response(r):
            return_folder = r.json()
            return self._json_to_folder(self.account, return_folder)

    def get_subfolders(self):
        """Retrieve all child Folders inside of this Folder.

        Raises:
            AuthError: Raised if Outlook returns a 401, generally caused by an invalid or expired access token.

        Returns:
            List[:class:`Folder <outlook_oauth2.core.folder.Folder>`]
        """
        headers = self.headers
        endpoint = 'https://outlook.office.com/api/v2.0' + self.account.route + 'MailFolders/' + self.id + '/childfolders'

        r = requests.get(endpoint, headers=headers)

        if check_response(r):
            return self._json_to_folders(self.account, r.json())

    def delete(self):
        """Deletes this Folder.

        Raises:
            AuthError: Raised if Outlook returns a 401, generally caused by an invalid or expired access token.

        """
        headers = self.headers
        endpoint = 'https://outlook.office.com/api/v2.0' + self.account.route + 'MailFolders/' + self.id

        r = requests.delete(endpoint, headers=headers)

        check_response(r)

    def move_into(self, destination_folder):
        # type: (Folder) -> Folder
        """Move the Folder into a different folder.

        This makes the Folder provided a child folder of the destination_folder.

        Raises:
            AuthError: Raised if Outlook returns a 401, generally caused by an invalid or expired access token.

        Args:
            destination_folder: A :class:`Folder <outlook_oauth2.core.folder.Folder>` that should become the parent

        Returns:
            A new :class:`Folder <outlook_oauth2.core.folder.Folder>` that is now
            inside of the destination_folder.

        """
        headers = self.headers
        endpoint = 'https://outlook.office.com/api/v2.0' + self.account.route + 'MailFolders/' + self.id + '/move'
        payload = '{ "DestinationId": "' + destination_folder.id + '"}'

        r = requests.post(endpoint, headers=headers, data=payload)

        if check_response(r):
            return_folder = r.json()
            return self._json_to_folder(self.account, return_folder)

    def copy_into(self, destination_folder):
        # type: (Folder) -> Folder
        """Copies the Folder into the provided destination folder.

        Raises:
            AuthError: Raised if Outlook returns a 401, generally caused by an invalid or expired access token.

        Args:
            destination_folder: The Folder that this Folder should be copied to.

        Returns:
            A new :class:`Folder <outlook_oauth2.core.folder.Folder>` representing the newly created folder.

        """
        headers = self.headers
        endpoint = 'https://outlook.office.com/api/v2.0' + self.account.route + 'MailFolders/' + self.id + '/copy'
        payload = '{ "DestinationId": "' + destination_folder.id + '"}'

        r = requests.post(endpoint, headers=headers, data=payload)

        if check_response(r):
            return_folder = r.json()
            return self._json_to_folder(self.account, return_folder)

    def create_child_folder(self, folder_name):
        """Creates a child folder within the Folder it is called from and returns the new Folder object.

        Args:
            folder_name: The name of the folder to create

        Returns: :class:`Folder <outlook_oauth2.core.folder.Folder>`
        """
        headers = self.headers
        endpoint = 'https://outlook.office.com/api/v2.0' + self.account.route + 'MailFolders/' + self.id + '/childfolders'
        payload = '{ "DisplayName": "' + folder_name + '"}'

        r = requests.post(endpoint, headers=headers, data=payload)

        if check_response(r):
            return_folder = r.json()
            return self._json_to_folder(self.account, return_folder)

    def messages(self):
        """ Retrieves the messages in this Folder,
        returning a list of :class:`Messages <outlook_oauth2.core.message.Message>`."""
        headers = self.headers
        r = requests.get('https://outlook.office.com/api/v2.0' + self.account.route + 'MailFolders/' + self.id + '/messages', headers=headers)
        check_response(r)
        from outlook_oauth2.core.message import Message
        return Message._json_to_messages(self.account, r.json())



# Authorization and misc functions
import json
import logging

from datetime import datetime

import requests

from outlook_oauth2.core.contact import Contact
from outlook_oauth2.core.message import Message
from outlook_oauth2.core.folder import Folder
from outlook_oauth2.internal.utils import check_response

log = logging.getLogger('outlook_oauth2')
__all__ = ['Mailbox']


class Mailbox(object):
    """Sets up access to Outlook account for all methods & classes.

    Attributes:
        access_token: A string OAuth token from Outlook allowing access to a user's account
    """

    def __init__(self, account, email_address=None):
        self.account = account
        self._auto_reply = None
        self._contact_overrides = None

        self.email_address = email_address

        if (self.email_address is not None) and (self.email_address != 'me'):
            self.route = 'users/' + self.email_address
        else:
            self.route = 'me'

    @property
    def _headers(self):
        return {"Authorization": "Bearer " + str(self.account.access_token), "Content-Type": "application/json"}

    @property
    def auto_reply_message(self):
        """ The account's Internal auto reply message. Setting the value will change the auto reply message of the
         account, automatically setting the status to enabled (but not altering the schedule). """
        if self._auto_reply is None:
            r = requests.get('https://outlook.office.com/api/v2.0/' + self.route + '/MailboxSettings/AutomaticRepliesSetting',
                             headers=self._headers)
            check_response(r)
            self._auto_reply = r.json().get('InternalReplyMessage')# Authorization and misc functions


        return self._auto_reply

    @auto_reply_message.setter
    def auto_reply_message(self, value):
        self.set_auto_reply(value)

    @property
    def contact_overrides(self):
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.route + '/InferenceClassification/Overrides'

        if self._contact_overrides is None:
            r = requests.get(endpoint, headers=self._headers)

            check_response(r)

            self._contact_overrides = Contact._json_to_contacts(r.json())

        return self._contact_overrides

    class AutoReplyAudience(object):
        INTERNAL_ONLY = 'None'
        CONTACTS_ONLY = 'ContactsOnly'
        ALL = 'All'

    class AutoReplyStatus(object):
        DISABLED = 'Disabled'
        ALWAYS_ENABLED = 'AlwaysEnabled'
        SCHEDULED = 'Scheduled'

    def set_auto_reply(self, message, status=AutoReplyStatus.ALWAYS_ENABLED, start=None, end=None,
                       external_message=None, audience=AutoReplyAudience.ALL):
        # type: (str, OutlookAccount.AutoReplyStatus, datetime, datetime, str, OutlookAccount.AutoReplyAudience) -> None
        """ Set an automatic reply for the account.
        Args:
            message (str): The message to be sent in replies. If external_message is provided this is the message sent
            to internal recipients
            status (OutlookAccount.AutoReplyStatus): Whether the auto-reply should be always enabled, scheduled, or
            disabled. You can use :class:`AutoReplyStatus <outlook_oauth2.core.main.OutlookAccount.AutoReplyStatus>` to
            provide the value. Defaults to ALWAYS_ENABLED.
            start (datetime): If status is set to SCHEDULED, this is when the replies will start being sent.
            end (datetime): If status is set to SCHEDULED, this is when the replies will stop being sent.
            external_message (str): If provided, this message will be sent to external recipients.
            audience (OutlookAccount.AutoReplyAudience): Whether replies should be sent to everyone, contacts only,
            or internal recipients only. You can use
            :class:`AutoReplyAudience <outlook_oauth2.core.main.OutlookAccount.AutoReplyAudience>` to provide the value.

        """

        start_is_none = start is None
        end_is_none = end is None

        if (not start_is_none and end_is_none) or (start_is_none and not end_is_none):
            raise ValueError('Start and End not must both either be None or datetimes')

        start_is_datetime = isinstance(start, datetime)
        end_is_datetime = isinstance(end, datetime)

        if not start_is_datetime and not start_is_none or not end_is_datetime and not end_is_none:
            raise ValueError('Start and End must both either be None or datetimes')

        request_data = dict(Status=status, ExternalAudience=audience)

        # Outlook requires both an internal and external message. For convenience, outlook_oauth2 allows only one message
        # and uses that as the external message if none is provided
        if external_message is None:
            external_message = message

        request_data.update(InternalReplyMessage=message, ExternalReplyMessage=external_message)

        if not start_is_none and not end_is_none:
            request_data.update(ScheduledStartDateTime=dict(DateTime=str(start)))
            request_data.update(ScheduledEndDateTime=dict(DateTime=str(end)))

        data = {
            "@odata.context": "https://outlook.office.com/api/v2.0/$metadata#Me/MailboxSettings",
            "AutomaticRepliesSetting": request_data
        }

        requests.patch('https://outlook.office.com/api/v2.0/' + self.route + '/MailboxSettings',
                       headers=self._headers, data=json.dumps(data))

        self._auto_reply = message

    def get_message(self, message_id):
        """Gets message matching provided id.

         the Outlook email matching the provided message_id.

        Args:
            message_id: A string for the intended message, provided by Outlook

        Returns:
            :class:`Message <outlook_oauth2.core.message.Message>`

        """
        r = requests.get('https://outlook.office.com/api/v2.0/' + self.route + '/messages/' + message_id, headers=self._headers)
        check_response(r)
        return Message._json_to_message(self, r.json())

    def get_messages(self, page=0):
        """Get first 10 messages in account, across all folders.

        Keyword Args:
            page (int): Integer representing the 'page' of results to fetch

        Returns:
            List[:class:`Message <outlook_oauth2.core.message.Message>`]

        """
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.route + '/messages'
        if page > 0:
            endpoint = endpoint + '/?%24skip=' + str(page) + '0'

        log.debug('Getting messages from endpoint: {} with Headers: {}'.format(endpoint, self._headers))

        r = requests.get(endpoint, headers=self._headers)

        check_response(r)

        return Message._json_to_messages(self, r.json())

    def inbox(self):
        """ first ten messages in account's inbox.

        Returns:
            List[:class:`Message <outlook_oauth2.core.message.Message>`]

        """
        return self._get_messages_from_folder_name('Inbox')

    def new_email(self, body='', subject='', to=list):
        """Creates a :class:`Message <outlook_oauth2.core.message.Message>` object.

        Keyword Args:
            body (str): The body of the email
            subject (str): The subject of the email
            to (List[Contact]): A list of recipients to email

        Returns:
            :class:`Message <outlook_oauth2.core.message.Message>`

        """
        return Message(self, body, subject, to)

    def send_email(self, body=None, subject=None, to=list, cc=None, bcc=None,
                   send_as=None, attachments=None):
        """Sends an email in one method, a shortcut for creating an instance of
        :class:`Message <outlook_oauth2.core.message.Message>` .

        Args:
            body (str): The body of the email
            subject (str): The subject of the email
            to (list): A list of :class:`Contacts <outlook_oauth2.core.contact.Contact>`
            cc (list): A list of :class:`Contacts <outlook_oauth2.core.contact.Contact>` which will be added to the
                'Carbon Copy' line
            bcc (list): A list of :class:`Contacts <outlook_oauth2.core.contact.Contact>` while be blindly added to the email
            send_as (Contact): A :class:`Contact <outlook_oauth2.core.contact.Contact>` whose email the OutlookAccount
                has access to
            attachments (list): A list of dictionaries with two parts
                [1] 'name' - a string which will become the file's name
                [2] 'bytes' - the bytes of the file.

        """
        email = Message(self, body, subject, to, cc=cc, bcc=bcc, sender=send_as,)

        if attachments is not None:
            for attachment in attachments:
                email.attach(attachment.get('bytes'), attachment.get('name'))

        email.send()

    def sent_messages(self):
        """ last ten sent messages.

        Returns:
            List[:class:`Message <outlook_oauth2.core.message.Message>`]

        """
        return self._get_messages_from_folder_name('SentItems')

    def deleted_messages(self):
        """ last ten deleted messages.

        Returns:
            List[:class:`Message <outlook_oauth2.core.message.Message>` ]

        """
        return self._get_messages_from_folder_name('DeletedItems')

    def draft_messages(self):
        """ last ten draft messages.

        Returns:
            List[:class:`Message <outlook_oauth2.core.message.Message>`]

        """
        return self._get_messages_from_folder_name('Drafts')

    def get_folders(self):
        """ Returns a list of all folders for this account

            Returns:
                List[:class:`Folder <outlook_oauth2.core.folder.Folder>`]
        """
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.route + '/MailFolders/'

        r = requests.get(endpoint, headers=self._headers)

        if check_response(r):
            return Folder._json_to_folders(self, r.json())

    def get_folder_by_id(self, folder_id):
        """ Retrieve a Folder by its Outlook ID

        Args:
            folder_id: The ID of the :class:`Folder <outlook_oauth2.core.folder.Folder>` to retrieve

        Returns: :class:`Folder <outlook_oauth2.core.folder.Folder>`

        """
        endpoint = 'https://outlook.office.com/api/v2.0/' + self.route + '/MailFolders/' + folder_id

        r = requests.get(endpoint, headers=self._headers)

        check_response(r)
        return_folder = r.json()
        return Folder._json_to_folder(self, return_folder)

    def _get_messages_from_folder_name(self, folder_name):
        """ Retrieves all messages from a folder, specified by its name. This only works with "Well Known" folders,
        such as 'Inbox' or 'Drafts'.

        Args:
            folder_name (str): The name of the folder to retrieve

        Returns: List[:class:`Message <outlook_oauth2.core.message.Message>` ]

        """
        r = requests.get('https://outlook.office.com/api/v2.0/' + self.route + '/MailFolders/' + folder_name + '/messages',
                         headers=self._headers)
        check_response(r)
        return Message._json_to_messages(self, r.json())



from base64 import b64decode

from datetime import datetime
from dateutil import parser

__all__ = ['Attachment']


class Attachment(object):
    def __init__(self, name, content, outlook_id=None, size=None, last_modified=None, content_type=None):
        # type: (str, str, str, int, datetime, str) -> None
        self.name = name

        self._content = content
        self.bytes = b64decode(content)

        self.outlook_id = outlook_id
        self.size = size
        self.last_modified = last_modified
        self.content_type = content_type

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    @classmethod
    def json_to_attachment(cls, account, api_json):
        outlook_id = api_json.get('Id')
        name = api_json.get('Name')

        content = api_json.get('ContentBytes', None)
        size = api_json.get('Size', None)
        content_type = api_json.get('ContentType', None)

        last_modified = api_json.get('LastModifiedDateTime', None)
        if last_modified is not None:
            parser.parse(last_modified)

        return Attachment(name, outlook_id=outlook_id, content=content, size=size,
                          content_type=content_type, last_modified=last_modified)

    @classmethod
    def json_to_attachments(cls, account, api_json):
        return [cls.json_to_attachment(account, value) for value in api_json['value']]

    def api_representation(self):
        """ Used for uploading attachments - less information is required than what we receive from the API """
        return {'@odata.type': '#Microsoft.OutlookServices.FileAttachment', 'Name': self.name,
                'ContentBytes': self._content}



import setuptools

from outlook_oauth2 import __release__

setuptools.setup(
    name='outlook_oauth2',
    version=__release__,
    packages=['outlook_oauth2', 'outlook_oauth2.internal', 'outlook_oauth2.core'],
    url='https://gitlab.com/alexbishop/pyOutlook',
    license='MIT',
    author='Alex Bishop',
    author_email='alexbishop1234@gmail.cpm',
    description='This is a fork of pyOutlook (https://pypi.python.org/pypi/pyOutlook) with support for multiple mailboxes.',
    install_requires=['requests', 'python-dateutil'],
    keywords='outlook office365 microsoft email'
)
